import os
import pathlib
import sys

# Import the C++ extension module and expose all its symbols
try:
    from . import _hepevd_impl

    # Bring all symbols into the HepEVD namespace
    for name in dir(_hepevd_impl):
        if not name.startswith('__'):
            globals()[name] = getattr(_hepevd_impl, name)

    # Keep a reference to the module
    globals()['_hepevd_impl'] = _hepevd_impl

except ImportError as e:
    # Simple error reporting is sufficient now that we know the issue
    print(f"ERROR: Failed to import HepEVD C++ extension: {e}", file=sys.stderr)
    raise

# Set up web folder path for the server
if not os.environ.get("HEP_EVD_WEB_FOLDER"):
    web_path = (pathlib.Path(__file__).parent / "web").resolve()
    os.environ["HEP_EVD_WEB_FOLDER"] = str(web_path)

# Clean up imports
del os
del pathlib
del sys


import pytest
import gc
import itertools
import re
import sys

try:
    import numpy as np
    from numpy.testing import assert_array_equal
    import test_eigen_ext as t
    def needs_numpy_and_eigen(x):
        return x
except:
    needs_numpy_and_eigen = pytest.mark.skip(reason="NumPy and Eigen are required")


@needs_numpy_and_eigen
def test01_vector_fixed():
    a  = np.array([1, 2, 3],    dtype=np.int32)
    b  = np.array([0, 1, 2],    dtype=np.int32)
    c  = np.array([1, 3, 5],    dtype=np.int32)
    x  = np.array([1, 3, 5, 6], dtype=np.int32)
    af = np.float32(a)
    bf = np.float32(b)

    assert_array_equal(t.addV3i(a, b), c)
    assert_array_equal(t.addR3i(a, b), c)
    assert_array_equal(t.addRefCnstV3i(a, b), c)
    assert_array_equal(t.addRefCnstR3i(a, b), c)
    assert_array_equal(t.addA3i(a, b), c)
    assert_array_equal(t.addA3i_retExpr(a, b), c)

    # Implicit conversion supported for first argument
    assert_array_equal(t.addV3i(af, b), c)
    assert_array_equal(t.addR3i(af, b), c)
    assert_array_equal(t.addRefCnstV3i(af, b), c)
    assert_array_equal(t.addRefCnstR3i(af, b), c)
    assert_array_equal(t.addA3i(af, b), c)

    # But not the second one
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.addV3i(a, bf)
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.addR3i(a, bf)
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.addRefCnstV3i(a, bf)
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.addRefCnstR3i(a, bf)
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.addA3i(a, bf)

    # Catch size errors
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.addV3i(x, b)
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.addR3i(x, b)
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.addRefCnstV3i(x, b)
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.addA3i(x, b)


@needs_numpy_and_eigen
def test02_vector_dynamic():
    a  = np.array([1, 2, 3], dtype=np.int32)
    b  = np.array([0, 1, 2], dtype=np.int32)
    c  = np.array([1, 3, 5], dtype=np.int32)
    x  = np.arange(10000, dtype=np.int32)
    af = np.float32(a)

    # Check call with dynamically sized arrays
    assert_array_equal(t.addVXi(a, b), c)

    # Implicit conversion
    assert_array_equal(t.addVXi(af, b), c)

    # Try with a big array. This will move the result to avoid a copy
    assert_array_equal(t.addVXi(x, x), 2*x)


@needs_numpy_and_eigen
def test03_update_map():
    a = np.array([1, 2, 3], dtype=np.int32)
    b = np.array([1, 2, 123], dtype=np.int32)
    c = a.copy()
    t.updateRefV3i(c)
    assert_array_equal(c, b)

    c = a.copy()
    t.updateRefV3i_nc(c)
    assert_array_equal(c, b)

    c = a.copy()
    t.updateRefVXi(c)
    assert_array_equal(c, b)

    c = a.copy()
    t.updateRefVXi_nc(c)
    assert_array_equal(c, b)

    c = np.float32(a)
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.updateRefV3i(c)

    c = np.float32(a)
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.updateRefV3i_nc(c)

    c = np.float32(a)
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.updateRefVXi(c)

    c = np.float32(a)
    with pytest.raises(TypeError, match='incompatible function arguments'):
        t.updateRefVXi_nc(c)



@needs_numpy_and_eigen
def test04_matrix():
    A = np.vander((1, 2, 3, 4,))
    At = A.T
    assert A.flags['C_CONTIGUOUS']
    assert At.flags['F_CONTIGUOUS']
    base = np.zeros((A.shape[0] * 2, A.shape[1] * 2), A.dtype)
    base[::2, ::2] = A
    Av = base[-2::-2, -2::-2]
    assert Av.base is base
    Avt = Av.T
    assert Avt.base is base
    matrices = A, At, Av, Avt
    for addM in (t.addM4uCC, t.addM4uRR, t.addM4uCR, t.addM4uRC,
                 t.addMXuCC, t.addMXuRR, t.addMXuCR, t.addMXuRC):
        for left, right in itertools.product(matrices, matrices):
            assert_array_equal(addM(left, right), left + right)


@needs_numpy_and_eigen
@pytest.mark.parametrize("rowStart", (0, 1))
@pytest.mark.parametrize("colStart", (0, 2))
@pytest.mark.parametrize("rowStep", (1, 2, -2))
@pytest.mark.parametrize("colStep", (1, 3, -3))
@pytest.mark.parametrize("transpose", (False, True))
def test05_matrix_large_nonsymm(rowStart, colStart, rowStep, colStep, transpose):
    A = np.uint32(np.vander(np.arange(80)))
    if rowStep < 0:
        rowStart = -rowStart - 1
    if colStep < 0:
        colStart = -colStart - 1
    A = A[rowStart::rowStep, colStart::colStep]
    if transpose:
        A = A.T
    A2 = A + A
    assert_array_equal(t.addMXuCC(A, A), A2)
    assert_array_equal(t.addMXuRR(A, A), A2)
    assert_array_equal(t.addMXuCR(A, A), A2)
    assert_array_equal(t.addMXuRC(A, A), A2)
    assert_array_equal(t.addDRefMXuCC_nc(A, A), A2)
    assert_array_equal(t.addDRefMXuRR_nc(A, A), A2)
    if A.flags['C_CONTIGUOUS']:
        assert_array_equal(t.addMapMXuRR(A, A), A2)
        assert_array_equal(t.addMapCnstMXuRR(A, A), A2)
    else:
        with pytest.raises(TypeError, match="incompatible function arguments"):
            t.addMapMXuRR(A, A)
        with pytest.raises(TypeError, match="incompatible function arguments"):
            t.addMapCnstMXuRR(A, A)

    assert_array_equal(t.addRefCnstMXuRR(A, A), A2)
    assert_array_equal(t.addRefCnstMXuRR(A.view(np.int32), A), A2)
    assert_array_equal(t.addRefCnstMXuRR_nc(A, A), A2)
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.addRefCnstMXuRR_nc(A.view(np.int32), A)
    if A.strides[1] == A.itemsize:
        assert_array_equal(t.addRefMXuRR(A, A), A2)
    else:
        with pytest.raises(TypeError, match="incompatible function arguments"):
            t.addRefMXuRR(A, A)
    if A.flags['F_CONTIGUOUS']:
        assert_array_equal(t.addMapMXuCC(A, A), A2)
        assert_array_equal(t.addMapCnstMXuCC(A, A), A2)
    else:
        with pytest.raises(TypeError, match="incompatible function arguments"):
            t.addMapMXuCC(A, A)
        with pytest.raises(TypeError, match="incompatible function arguments"):
            t.addMapCnstMXuCC(A, A)

    assert_array_equal(t.addRefCnstMXuCC(A, A), A2)
    assert_array_equal(t.addRefCnstMXuCC(A.view(np.int32), A), A2)
    assert_array_equal(t.addRefCnstMXuCC_nc(A, A), A2)
    with pytest.raises(TypeError, match="incompatible function arguments"):
        t.addRefCnstMXuCC_nc(A.view(np.int32), A)
    if A.strides[0] == A.itemsize:
        assert_array_equal(t.addRefMXuCC(A, A), A2)
    else:
        with pytest.raises(TypeError, match="incompatible function arguments"):
            t.addRefMXuCC(A, A)
    A = np.ascontiguousarray(A)
    assert A.flags['C_CONTIGUOUS']
    assert_array_equal(t.addMXuRR_nc(A, A), A2)
    A = np.asfortranarray(A)
    assert A.flags['F_CONTIGUOUS']
    assert_array_equal(t.addMXuCC_nc(A, A), A2)


@needs_numpy_and_eigen
def test06_map():
    b = t.Buffer()
    m = b.map()
    dm = b.dmap()
    for i in range(10):
        for j in range(3):
            m[i, j] = i*3+j
    for i in range(10):
        for j in range(3):
            assert dm[i, j] == i*3+j
    del dm
    del b
    gc.collect()
    gc.collect()
    for i in range(10):
        for j in range(3):
            assert m[i, j] == i*3+j


@needs_numpy_and_eigen
def test07_mutate_arg():
    A = np.uint32(np.vander(np.arange(10)))
    A2 = A.copy()
    t.mutate_DRefMXuC(A)
    assert_array_equal(A, 2*A2)


@needs_numpy_and_eigen
def test08_sparse():
    pytest.importorskip("scipy")
    import scipy.sparse

    # no isinstance here because we want strict type equivalence
    assert type(t.sparse_r()) is scipy.sparse.csr_matrix
    assert type(t.sparse_c()) is scipy.sparse.csc_matrix
    assert type(t.sparse_copy_r(t.sparse_r())) is scipy.sparse.csr_matrix
    assert type(t.sparse_copy_c(t.sparse_c())) is scipy.sparse.csc_matrix
    assert type(t.sparse_copy_r(t.sparse_c())) is scipy.sparse.csr_matrix
    assert type(t.sparse_copy_c(t.sparse_r())) is scipy.sparse.csc_matrix

    def assert_sparse_equal_ref(sparse_mat):
        ref = np.array(
            [
                [0.0, 3, 0, 0, 0, 11],
                [22, 0, 0, 0, 17, 11],
                [7, 5, 0, 1, 0, 11],
                [0, 0, 0, 0, 0, 11],
                [0, 0, 14, 0, 8, 11],
            ]
        )
        assert_array_equal(sparse_mat.toarray(), ref)

    assert_sparse_equal_ref(t.sparse_r())
    assert_sparse_equal_ref(t.sparse_c())
    assert_sparse_equal_ref(t.sparse_copy_r(t.sparse_r()))
    assert_sparse_equal_ref(t.sparse_copy_c(t.sparse_c()))
    assert_sparse_equal_ref(t.sparse_copy_r(t.sparse_c()))
    assert_sparse_equal_ref(t.sparse_copy_c(t.sparse_r()))


@needs_numpy_and_eigen
def test09_sparse_failures():
    pytest.importorskip("scipy")
    import scipy

    with pytest.raises(
        ValueError,
        match=re.escape(
            "nanobind: unable to return an Eigen sparse matrix that is not in a compressed format. Please call `.makeCompressed()` before returning the value on the C++ end."
        ),
    ):
        t.sparse_r_uncompressed()

    csr_matrix = scipy.sparse.csr_matrix
    scipy.sparse.csr_matrix = None
    with pytest.raises(TypeError, match=re.escape("'NoneType' object is not callable")):
        t.sparse_r()

    del scipy.sparse.csr_matrix
    with pytest.raises(
        AttributeError,
        match=re.escape("module 'scipy.sparse' has no attribute 'csr_matrix'"),
    ):
        t.sparse_r()

    sys_path = sys.path
    sys.path = []
    del sys.modules["scipy"]
    with pytest.raises(ModuleNotFoundError, match=re.escape("No module named 'scipy'")):
        t.sparse_r()

    # undo sabotage of the module
    sys.path = sys_path
    scipy.sparse.csr_matrix = csr_matrix

@needs_numpy_and_eigen
def test10_eigen_scalar_default():
    x = t.default_arg()
    assert x==0

@needs_numpy_and_eigen
def test11_prop():
    for j in range(3):
        c = t.ClassWithEigenMember()
        ref = np.ones((2, 2))
        if j == 0:
            c.member = ref

        for i in range(2):
            member = c.member
            if j == 2 and i == 0:
                member[0, 0] = 10
                ref[0, 0] = 10
            assert_array_equal(member, ref)
            del member
            gc.collect()
            gc.collect()

        member = c.member
        del c
        gc.collect()
        gc.collect()
        assert_array_equal(member, ref)

@needs_numpy_and_eigen
def test12_cast():
    vec = np.arange(1000, dtype=np.int32)
    vec2 = vec[::2]
    vecf = np.float32(vec)
    assert_array_equal(t.castToMapVXi(vec), vec)
    assert_array_equal(t.castToMapCnstVXi(vec), vec)
    assert_array_equal(t.castToRefVXi(vec), vec)
    assert_array_equal(t.castToRefCnstVXi(vec), vec)
    assert t.castToMapVXi(vec).flags.writeable
    assert not t.castToMapCnstVXi(vec).flags.writeable
    assert_array_equal(t.castToDRefCnstVXi(vec), vec)
    for v in vec2, vecf:
        with pytest.raises(RuntimeError, match="bad[_ ]cast"):
            t.castToMapVXi(v)
        with pytest.raises(RuntimeError, match="bad[_ ]cast"):
            t.castToRefVXi(v)
        assert_array_equal(t.castToRefCnstVXi(v), v)
    assert_array_equal(t.castToDRefCnstVXi(vec2), vec2)
    with pytest.raises(RuntimeError, match="bad[_ ]cast"):
        t.castToDRefCnstVXi(vecf)
    for v in vec, vec2, vecf:
        with pytest.raises(RuntimeError, match='bad[_ ]cast'):
            t.castToRef03CnstVXi(v)

@needs_numpy_and_eigen
def test13_mutate_python():
    class Derived(t.Base):
        def modRefData(self, input):
            input[0] = 3.0

        def modRefDataConst(self, input):
            input[0] = 3.0

    vecRef = np.array([3.0, 2.0])
    der = Derived()
    assert_array_equal(t.modifyRef(der), vecRef)
    with pytest.raises(ValueError):
        t.modifyRefConst(der)


@needs_numpy_and_eigen
def test14_single_element():
    a = np.array([[1]], dtype=np.uint32)
    assert a.ndim == 2 and a.shape == (1, 1)
    t.addMXuCC(a, a)



import pytest
import platform

import test_bind_vector_ext as t

def test01_vector_int(capfd):
    v_int = t.VectorInt([0, 0])
    assert len(v_int) == 2
    assert bool(v_int) is True

    # test construction from a generator
    v_int1 = t.VectorInt(x for x in range(5))
    assert t.VectorInt(v_int1) == t.VectorInt([0, 1, 2, 3, 4])
    assert repr(v_int1) == "test_bind_vector_ext.VectorInt([0, 1, 2, 3, 4])"

    v_int2 = t.VectorInt([0, 0])
    assert v_int == v_int2
    v_int2[1] = 2
    v_int2[1] -= 1
    assert v_int2[1] == 1
    assert v_int != v_int2

    v_int2.append(2)
    v_int2.insert(0, 1)
    v_int2.insert(0, 2)
    v_int2.insert(0, 3)
    v_int2.insert(6, 3)
    with pytest.raises(IndexError):
        v_int2.insert(8, 4)

    v_int.append(99)
    v_int2[2:-2] = v_int
    assert v_int2 == t.VectorInt([3, 2, 0, 0, 99, 2, 3])
    del v_int2[1:3]
    assert v_int2 == t.VectorInt([3, 0, 99, 2, 3])
    del v_int2[0]
    assert v_int2 == t.VectorInt([0, 99, 2, 3])

    v_int2.extend(t.VectorInt([4, 5]))
    assert v_int2 == t.VectorInt([0, 99, 2, 3, 4, 5])

    v_int2.extend([6, 7])
    assert v_int2 == t.VectorInt([0, 99, 2, 3, 4, 5, 6, 7])

    # test error handling, and that the vector is unchanged
    with pytest.raises(TypeError):
        v_int2.extend([8, "a"])

    captured = capfd.readouterr().err.strip()
    ref = "nanobind: implicit conversion from type 'list' to type 'test_bind_vector_ext.VectorInt' failed!"

    # Work around Pytest-related flakiness (https://github.com/pytest-dev/pytest/issues/10843)
    if platform.system() == 'Windows':
        assert captured == ref or captured == ''
    else:
        assert captured == ref

    assert v_int2 == t.VectorInt([0, 99, 2, 3, 4, 5, 6, 7])

    # test extending from a generator
    v_int2.extend(x for x in range(5))
    assert v_int2 == t.VectorInt([0, 99, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 4])

    # Test count feature
    assert v_int2.count(2) == 2
    assert v_int2.count(5) == 1
    assert v_int2.count(8) == 0
    assert 2 in v_int2
    assert 5 in v_int2
    assert 8 not in v_int2

    # test negative indexing
    assert v_int2[-1] == 4

    # insert with negative index
    v_int2.insert(-1, 88)
    assert v_int2 == t.VectorInt([0, 99, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 88, 4])

    # delete negative index
    del v_int2[-1]
    assert v_int2 == t.VectorInt([0, 99, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3, 88])

    assert v_int2.pop() == 88
    assert v_int2 == t.VectorInt([0, 99, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3])

    assert v_int2.pop(1) == 99
    assert v_int2 == t.VectorInt([0, 2, 3, 4, 5, 6, 7, 0, 1, 2, 3])

    v_int2.clear()
    assert len(v_int2) == 0


def test02_vector_bool():
    vv_c = t.VectorBool()
    for i in range(9):
        vv_c.append(i % 2 == 0)
    for i in range(9):
        assert vv_c[i] == (i % 2 == 0)
    assert vv_c.count(True) == 5
    assert vv_c.count(False) == 4


def test03_vector_custom():
    v_a = t.VectorEl()
    v_a.append(t.El(1))
    v_a.append(t.El(2))
    assert len(v_a) == 2 and v_a[0].a == 1 and v_a[1].a == 2

    vv_a = t.VectorVectorEl()
    vv_a.append(v_a)
    v_b = vv_a[0]
    assert len(v_b) == 2 and v_b[0].a == 1 and v_b[1].a == 2


def test04_vector_slicing():
    l1 = list(range(100))
    l2 = t.VectorInt(l1)

    def check_same(s):
        assert l1[s] == l2[s]

    def check_del(s):
        l1c = type(l1)(l1)
        l2c = type(l2)(l2)
        del l1c[s]
        del l2c[s]
        l2c = list(l2c)
        assert l1c == l2c

    check_same(slice(1, 13, 4))
    check_same(slice(1, 14, 4))
    check_same(slice(10, 2000, 1))
    check_same(slice(200, 10, 1))
    check_same(slice(200, 10, -1))
    check_same(slice(200, 10, -3))
    check_del(slice(1, 13, 4))
    check_del(slice(1, 14, 4))
    check_del(slice(10, 2000, 1))
    check_del(slice(200, 10, 1))
    check_del(slice(200, 10, -1))
    check_del(slice(200, 10, -3))


def test05_vector_non_shared():
    v = t.VectorEl()
    v.append(t.El(1))
    v.append(t.El(2))

    v0, v1 = v[0], v[1]
    v0.a, v1.a = 100, 200

    assert v[0].a == 1
    assert v[1].a == 2

    # Check that elements accessed through the iterator *cannot* be modified
    q = next(iter(v))
    q.a = 5
    assert v[0].a == 1

def test06_vector_shared():
    v = t.VectorElShared()
    v.append(t.El(1))
    v.append(t.El(2))

    v0, v1 = v[0], v[1]
    v0.a, v1.a = 100, 200

    assert v[0].a == 100
    assert v[1].a == 200

    # Check that elements accessed through the iterator *can* be modified
    q = next(iter(v))
    q.a = 5
    assert v[0].a == 5


def test07_vector_noncopyable():
    vnc = t.get_vnc(5)
    for i in range(0, 5):
        assert vnc[i].value == i + 1

    for i, j in enumerate(vnc, start=1):
        assert j.value == i

    # Check that elements accessed through the iterator *can* be modified
    q = next(iter(vnc))
    q.value = 5
    assert vnc[0].value == 5



import test_enum_ext as t
import pytest

def test01_unsigned_enum():
    assert repr(t.Enum.A) == 'Enum.A'
    assert str(t.Enum.A) == 'Enum.A'
    assert repr(t.Enum.B) == 'Enum.B'
    assert str(t.Enum.B) == 'Enum.B'
    assert repr(t.Enum.C) == 'Enum.C'
    assert str(t.Enum.C) == 'Enum.C'
    assert t.Enum.A.name == 'A'
    assert t.Enum.B.name == 'B'
    assert t.Enum.C.name == 'C'
    assert t.Enum.A.__name__ == 'A'
    assert t.Enum.B.__name__ == 'B'
    assert t.Enum.C.__name__ == 'C'

    assert t.Enum.__doc__ == 'enum-level docstring'

    assert t.Enum.A.__doc__ == 'Value A'
    assert t.Enum.B.__doc__ == 'Value B'
    assert t.Enum.C.__doc__ == 'Value C'
    assert t.Enum.A.value == 0
    assert t.Enum.B.value == 1
    assert t.Enum.C.value == 0xffffffff
    assert t.Enum(0) is t.Enum.A
    assert t.Enum(1) is t.Enum.B
    assert t.Enum(0xffffffff) is t.Enum.C
    assert t.Enum(t.Enum.A) is t.Enum.A
    assert t.Enum(t.Enum.B) is t.Enum.B
    assert t.Enum(t.Enum.C) == t.Enum.C
    assert t.from_enum(t.Enum.A) == 0
    assert t.from_enum(t.Enum.B) == 1
    assert t.from_enum(t.Enum.C) == 0xffffffff
    assert t.to_enum(0).__name__ == 'A'
    assert t.to_enum(0) == t.Enum.A
    assert t.to_enum(1) == t.Enum.B
    assert t.to_enum(0xffffffff) == t.Enum.C

    with pytest.raises(ValueError) as excinfo:
        t.to_enum(5)
    assert '5 is not a valid Enum' in str(excinfo.value)

    with pytest.raises(ValueError) as excinfo:
        t.Enum(123)
    assert '123 is not a valid Enum' in str(excinfo.value)

    assert t.from_enum_implicit(0) == 0

    with pytest.raises(TypeError):
        t.from_enum_implicit(123)


def test02_signed_enum():
    assert repr(t.SEnum.A) == 'SEnum.A'
    assert repr(t.SEnum.B) == 'SEnum.B'
    assert repr(t.SEnum.C) == 'SEnum.C'
    assert str(t.SEnum.A) == 'SEnum.A'
    assert str(t.SEnum.B) == 'SEnum.B'
    assert str(t.SEnum.C) == 'SEnum.C'
    assert int(t.SEnum.A) == 0
    assert int(t.SEnum.B) == 1
    assert int(t.SEnum.C) == -1
    assert t.SEnum.A.value == 0
    assert t.SEnum.B.value == 1
    assert t.SEnum.C.value == -1
    assert t.SEnum(0) is t.SEnum.A
    assert t.SEnum(1) is t.SEnum.B
    assert t.SEnum(-1) is t.SEnum.C
    assert t.from_enum(t.SEnum.A) == 0
    assert t.from_enum(t.SEnum.B) == 1
    assert t.from_enum(t.SEnum.C) == -1


def test03_enum_arithmetic():
    assert t.SEnum.B + 2 == 3
    assert t.SEnum.B + 2.5 == 3.5
    assert 2 + t.SEnum.B == 3
    assert 2.5 + t.SEnum.B == 3.5
    assert t.SEnum.B >> t.SEnum.B == 0
    assert t.SEnum.B << t.SEnum.B == 2
    assert -t.SEnum.B == -1 and -t.SEnum.C == 1
    assert t.SEnum.B & t.SEnum.B == 1
    assert t.SEnum.B & ~t.SEnum.B == 0

    with pytest.raises(TypeError, match="unsupported operand type"):
        t.Enum.B + 2
    with pytest.raises(TypeError, match="unsupported operand type"):
        t.SEnum.B - "1"
    with pytest.raises(TypeError, match="unsupported operand type"):
        t.SEnum.B >> 1.0


def test04_enum_export():
    assert t.Item1 is t.ClassicEnum.Item1 and t.Item1.value == 0
    assert t.Item2 is t.ClassicEnum.Item2 and t.Item2.value == 1

# test for issue #39
def test05_enum_property():
    w = t.EnumProperty()
    assert w.read_enum == t.Enum.A
    assert str(w.read_enum) == 'Enum.A'

def test08_enum_comparisons():
    assert int(t.SEnum.B) == 1
    for enum in (t.SEnum,):
        value = getattr(enum, "B")
        assert value != str(int(value))
        assert value != int(value) + 0.4
        assert value < int(value) + 0.4
        for i in (0, 0.5, 1, 1.5, 2):
            assert (value == i) == (int(value) == i)
            assert (value != i) == (int(value) != i)
            assert (value < i) == (int(value) < i)
            assert (value <= i) == (int(value) <= i)
            assert (value >= i) == (int(value) >= i)
            assert (value > i) == (int(value) > i)

            assert (i == value) == (i == int(value))
            assert (i != value) == (i != int(value))
            assert (i < value) == (i < int(value))
            assert (i <= value) == (i <= int(value))
            assert (i >= value) == (i >= int(value))
            assert (i > value) == (i > int(value))

        for unrelated in (None, "hello", "1"):
            assert value != unrelated and unrelated != value
            assert not (value == unrelated) and not (unrelated == value)
            with pytest.raises(TypeError):
                value < unrelated
            with pytest.raises(TypeError):
                unrelated < value

    # different enum types never compare equal ...
    assert t.Enum.B != t.SEnum.B and t.SEnum.B != t.Enum.B
    assert not (t.Enum.B == t.SEnum.B) and not (t.SEnum.B == t.Enum.B)
    assert t.Enum.B != t.SEnum.C and t.SEnum.C != t.Enum.B

def test09_enum_methods():
    assert t.Item1.my_value == 0 and t.Item2.my_value == 1
    assert t.Item1.get_value() == 0 and t.Item2.get_value() == 1
    assert t.Item1.foo() == t.Item1
    assert t.ClassicEnum.bar(t.Item1) == t.Item1



from setuptools import setup
import re
import os
import shutil


VERSION_REGEX = re.compile(
    r"^\s*#\s*define\s+NB_VERSION_([A-Z]+)\s+(.*)$", re.MULTILINE)

this_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(this_directory, "include/nanobind/nanobind.h")) as f:
    matches = dict(VERSION_REGEX.findall(f.read()))
    nanobind_version = "{MAJOR}.{MINOR}.{PATCH}".format(**matches)

long_description = '''
![nanobind logo](
https://github.com/wjakob/nanobind/raw/master/docs/images/logo.jpg?raw=True)

_nanobind_ is a small binding library that exposes C++ types in Python and
vice versa. It is reminiscent of
[Boost.Python](https://www.boost.org/doc/libs/1_64_0/libs/python/doc/html)
and [pybind11](http://github.com/pybind/pybind11) and uses near-identical
syntax. In contrast to these existing tools, nanobind is more efficient:
bindings compile in a shorter amount of time, produce smaller binaries, and
have better runtime performance.

More concretely,
[benchmarks](https://nanobind.readthedocs.io/en/latest/benchmark.html) show up
to **~4× faster** compile time, **~5× smaller** binaries, and **~10× lower**
runtime overheads compared to pybind11. nanobind also outperforms Cython in
important metrics (**3-12×** binary size reduction, **1.6-4×** compilation time
reduction, similar runtime performance).

Please see the following links for tutorial and reference documentation in
[HTML](https://nanobind.readthedocs.io/en/latest/) and
[PDF](https://nanobind.readthedocs.io/_/downloads/en/latest/pdf/) formats.
'''

from tempfile import TemporaryDirectory

with TemporaryDirectory() as temp_dir:
    base_dir = os.path.abspath(os.path.dirname(__file__))


    for name in ['include', 'ext', 'cmake', 'src']:
        shutil.copytree(os.path.join(base_dir, name),
                        os.path.join(temp_dir, name),
                        dirs_exist_ok=True)

    for fname in ['__init__.py', '__main__.py', 'stubgen.py']:
        shutil.move(os.path.join(temp_dir, 'src', fname),
                    os.path.join(temp_dir, fname))

    setup(
        name="nanobind",
        version=nanobind_version,
        author="Wenzel Jakob",
        author_email="wenzel.jakob@epfl.ch",
        description='nanobind: tiny and efficient C++/Python bindings',
        url="https://github.com/wjakob/nanobind",
        license="BSD",
        long_description=long_description,
        long_description_content_type='text/markdown',
        packages=['nanobind'],
        zip_safe=False,
        package_dir={'nanobind': temp_dir},
        package_data={'nanobind': [
            'include/nanobind/*.h',
            'include/nanobind/stl/*.h',
            'include/nanobind/stl/detail/*.h',
            'include/nanobind/eigen/*.h',
            'include/nanobind/intrusive/*.h',
            'include/nanobind/intrusive/*.inl',
            'ext/robin_map/include/tsl/robin_map.h',
            'ext/robin_map/include/tsl/robin_hash.h',
            'ext/robin_map/include/tsl/robin_growth_policy.h',
            'cmake/nanobind-config.cmake',
            'cmake/darwin-ld-cpython.sym',
            'cmake/darwin-ld-pypy.sym',
            'src/*.h',
            'src/*.cpp'
        ]}
    )



import sys
import test_classes_ext as t
import pytest
from common import skip_on_pypy, collect

# Some helper declaration to check types across different Python versions
if sys.version_info < (3, 9):
    TYPING_TYPE = "typing.Type"
else:
    TYPING_TYPE = "type"


def optional(arg: str, /) -> str:
    if sys.version_info < (3, 10):
        return "typing.Optional[" + arg + "]"
    else:
        return arg + " | " + "None"


@pytest.fixture
def clean():
    collect()
    t.reset()


def assert_stats(**kwargs):
    collect()
    for k, v in t.stats().items():
        fail = False
        if k in kwargs:
            if v != kwargs[k]:
                fail = True
        elif v != 0:
            fail = True
        if fail:
            raise Exception(f"Mismatch for key {k}: {t.stats()}")


def test01_signature():
    assert t.Struct.__init__.__doc__ == (
        "__init__(self) -> None\n" "__init__(self, arg: int, /) -> None"
    )

    assert t.Struct.value.__doc__ == "value(self) -> int"
    assert t.Struct.create_move.__doc__ == "create_move() -> test_classes_ext.Struct"
    assert t.Struct.set_value.__doc__ == "set_value(self, value: int) -> None"
    assert t.Struct().set_value.__doc__ == "set_value(self, value: int) -> None"
    assert t.Struct.__doc__ == "Some documentation"
    assert t.Struct.static_test.__doc__ == (
        "static_test(arg: int, /) -> int\n" "static_test(arg: float, /) -> int"
    )


def test02_static_overload():
    assert t.Struct.static_test(1) == 1
    assert t.Struct.static_test(1.0) == 2


def test03_instantiate(clean):
    s1: t.Struct = t.Struct()
    assert s1.value() == 5
    s2 = t.Struct(10)
    assert s2.value() == 10
    del s1
    del s2
    assert_stats(default_constructed=1, value_constructed=1, destructed=2)


def test04_double_init():
    s = t.Struct()
    with pytest.warns(
        RuntimeWarning,
        match="nanobind: attempted to initialize an already-initialized instance of type",
    ):
        with pytest.raises(TypeError):
            s.__init__(3)


def test05_rv_policy(clean):
    s = t.Struct()
    assert s.self() is s
    assert s.none() is None
    del s
    assert_stats(default_constructed=1, destructed=1)

    # ------

    t.reset()
    assert t.Struct.create_take().value() == 10
    assert_stats(value_constructed=1, destructed=1)

    # ------

    t.reset()
    assert t.Struct.create_move().value() == 11
    assert_stats(value_constructed=1, move_constructed=1, destructed=2)

    # ------

    t.reset()
    assert t.Struct.create_reference().value() == 12
    assert_stats()

    # ------

    t.reset()
    assert t.Struct.create_copy().value() == 12
    assert_stats(copy_constructed=1, destructed=1)


def test06_reference_internal(clean):
    s = t.PairStruct()
    s1 = s.s1
    s2 = s.s2
    del s
    assert_stats(default_constructed=2)
    assert s2.value() == 5
    del s2

    assert_stats(default_constructed=2)

    assert s1.value() == 5
    del s1
    assert_stats(default_constructed=2, destructed=2)

    # ----

    s = t.PairStruct()
    s1 = s.s1
    del s1
    del s


def test07_big():
    x = [t.Big() for i in range(1024)]
    x2 = [t.BigAligned() for i in range(1024)]


def test08_inheritance():
    dog = t.Dog("woof")
    cat = t.Cat("meow")
    assert dog.name() == "Dog"
    assert cat.name() == "Cat"
    assert dog.what() == "woof"
    assert cat.what() == "meow"
    assert isinstance(dog, t.Animal) and isinstance(dog, t.Dog)
    assert isinstance(cat, t.Animal) and isinstance(cat, t.Cat)
    assert t.go(dog) == "Dog says woof"
    assert t.go(cat) == "Cat says meow"

    assert t.animal_passthrough(dog) is dog
    assert t.animal_passthrough(cat) is cat
    assert t.dog_passthrough(dog) is dog

    with pytest.raises(TypeError):
        t.dog_passthrough(cat)


def test09_method_vectorcall():
    out = []

    def f(a, b, c, d, e):
        out.append((a, b, c, d, e))

    class MyClass:
        def f(self, a, b, c, d, e):
            self.out = (a, b, c, d, e)

    t.call_function(f)

    i = MyClass()
    t.call_method(i)
    assert out == [(1, 2, "hello", True, 4)]
    assert i.out == (1, 2, "hello", True, 4)


def test10_trampoline(clean):
    for _ in range(10):

        class Dachshund(t.Animal):
            def __init__(self):
                super().__init__()

            def name(self):
                return "Dachshund"

            def what(self):
                return "yap"

        d = Dachshund()
        for _ in range(10):
            assert t.go(d) == "Dachshund says yap"

        assert t.animal_passthrough(d) is d

    a = 0

    class GenericAnimal(t.Animal):
        def what(self):
            return "goo"

        def void_ret(self):
            nonlocal a
            a += 1

        def name(self):
            return "Generic" + super().name()

    ga = GenericAnimal()
    assert t.go(ga) == "GenericAnimal says goo"
    assert t.void_ret(ga) is None
    assert a == 1
    assert t.animal_passthrough(ga) is ga

    del ga
    del d

    assert_stats(default_constructed=11, destructed=11)

    class GenericDog(t.Dog):
        pass

    d = GenericDog("GenericDog")
    assert t.dog_passthrough(d) is d
    assert t.animal_passthrough(d) is d


def test11_trampoline_failures():
    class Incomplete(t.Animal):
        def __init__(self):
            super().__init__()

        def void_ret(self):
            raise TypeError("propagating an exception")

    d = Incomplete()
    with pytest.raises(RuntimeError) as excinfo:
        t.go(d)
    assert (
        "test_classes.Incomplete::what()'): tried to call a pure virtual function!"
        in str(excinfo.value)
    )

    with pytest.raises(TypeError) as excinfo:
        t.void_ret(d)
    assert "propagating an exception" in str(excinfo.value)

    class Incomplete2(t.Animal):
        def __init__(self):
            pass  # Missing call to super().__init__()

        def name(self):
            return "a"

        def what(self):
            return "b"

    with pytest.warns(
        RuntimeWarning,
        match="nanobind: attempted to access an uninitialized instance of type",
    ):
        with pytest.raises(TypeError) as excinfo:
            t.go(Incomplete2())
    assert "incompatible function arguments" in str(excinfo.value)


def test12_large_pointers():
    import struct

    for i in range(1, 10):
        c = t.i2p(i)
        assert isinstance(c, t.Foo)
        assert t.p2i(c) == i

    large = (1 << (struct.calcsize("P") * 8)) - 1
    for i in range(large - 10, large):
        c = t.i2p(i)
        assert isinstance(c, t.Foo)
        assert t.p2i(c) == i


def test13_implicitly_convertible():
    assert t.get_d.__doc__ == "get_d(arg: test_classes_ext.D, /) -> int"
    a = t.A(1)
    b = t.B(2)
    b2 = t.B2(3)
    c = t.C(4)
    i = 5

    with pytest.raises(TypeError) as excinfo:
        t.get_d(c)
    assert str(excinfo.value) == (
        "get_d(): incompatible function arguments. The following argument types are supported:\n"
        "    1. get_d(arg: test_classes_ext.D, /) -> int\n"
        "\n"
        "Invoked with types: test_classes_ext.C"
    )
    with pytest.raises(TypeError):
        t.get_optional_d(c)

    for obj, expected in ((a, 11), (b, 102), (b2, 103), (i, 10005)):
        assert t.get_d(obj) == expected
        assert t.get_optional_d(obj) == expected
        # The -1's here are because nb::cast() won't implicit-convert to a
        # pointer because it would dangle
        assert t.get_d_via_cast(obj) == (expected, -1, expected, -1)
        assert t.get_d_via_try_cast(obj) == (expected, -1, expected, -1)

    d = t.D(5)
    assert t.get_d(d) == 10005
    assert t.get_optional_d(d) == 10005
    assert t.get_d_via_cast(d) == (10005, 10005, 10005, 10005)
    assert t.get_d_via_try_cast(d) == (10005, 10005, 10005, 10005)

    assert t.get_optional_d(None) == -1
    assert t.get_d_via_cast(c) == (-1, -1, -1, -1)
    assert t.get_d_via_try_cast(c) == (-1, -1, -1, -1)


def test14_operators():
    a = t.Int(1)
    b = t.Int(2)
    assert repr(a + b) == "3"
    with pytest.raises(TypeError) as excinfo:
        assert repr(a - b) == "3"
    assert "unsupported operand type" in str(excinfo.value)
    assert repr(a - 2) == "-1"
    a += b
    assert repr(a) == "3"
    assert repr(b) == "2"

    assert a.__add__("test") is NotImplemented


def test15_keep_alive_nbtype(clean):
    t.reset()
    s = t.Struct()
    a = t.Dog("Rufus")
    assert t.keep_alive_arg(s, a) is a
    assert t.keep_alive_arg(s, a) is a
    del s
    assert_stats(default_constructed=1)
    del a
    assert_stats(default_constructed=1, destructed=1)

    t.reset()
    s = t.Struct()
    a = t.Dog("Rufus")
    assert t.keep_alive_ret(a, s) is s
    assert t.keep_alive_ret(a, s) is s
    del a
    assert_stats(default_constructed=1)
    del s
    assert_stats(default_constructed=1, destructed=1)


def test16_keep_alive_custom(clean):
    constructed = 0
    destructed = 0

    class Struct:
        def __init__(self):
            nonlocal constructed
            constructed += 1

        def __del__(self):
            nonlocal destructed
            destructed += 1

    class Struct2:
        def __init__(self):
            pass

    s = Struct()
    a = Struct2()
    assert t.keep_alive_arg(s, a) is a
    del s
    collect()
    assert constructed == 1 and destructed == 0
    del a
    collect()
    assert constructed == 1 and destructed == 1

    s = Struct()
    a = Struct2()
    assert t.keep_alive_ret(a, s) is s
    del a
    collect()
    assert constructed == 2 and destructed == 1
    del s
    collect()
    assert constructed == 2 and destructed == 2

    with pytest.raises(RuntimeError) as excinfo:
        s = Struct()
        x = 5
        t.keep_alive_ret(x, s)

    assert "nanobind::detail::keep_alive(): could not create a weak reference!" in str(
        excinfo.value
    )


def f():
    pass


class MyClass:
    def f(self):
        pass

    class NestedClass:
        def f(self):
            pass


def test17_name_qualname_module():
    # First, check what CPython does
    assert f.__module__ == "test_classes"
    assert f.__name__ == "f"
    assert f.__qualname__ == "f"
    assert MyClass.__name__ == "MyClass"
    assert MyClass.__qualname__ == "MyClass"
    assert MyClass.__module__ == "test_classes"
    assert MyClass.f.__name__ == "f"
    assert MyClass.f.__qualname__ == "MyClass.f"
    assert MyClass.f.__module__ == "test_classes"
    assert MyClass().f.__name__ == "f"
    assert MyClass().f.__qualname__ == "MyClass.f"
    assert MyClass().f.__module__ == "test_classes"
    assert MyClass.NestedClass.__name__ == "NestedClass"
    assert MyClass.NestedClass.__qualname__ == "MyClass.NestedClass"
    assert MyClass.NestedClass.__module__ == "test_classes"
    assert MyClass.NestedClass.f.__name__ == "f"
    assert MyClass.NestedClass.f.__qualname__ == "MyClass.NestedClass.f"
    assert MyClass.NestedClass.f.__module__ == "test_classes"

    # Now, check the extension module
    assert t.f.__module__ == "test_classes_ext"
    assert t.f.__name__ == "f"
    assert t.f.__qualname__ == "f"
    assert type(t.f).__module__ == "nanobind"
    assert type(t.f).__name__ == "nb_func"
    assert type(t.f).__qualname__ == "nb_func"
    assert t.MyClass.__name__ == "MyClass"
    assert t.MyClass.__qualname__ == "MyClass"
    assert t.MyClass.__module__ == "test_classes_ext"
    assert t.MyClass.f.__name__ == "f"
    assert t.MyClass.f.__qualname__ == "MyClass.f"
    assert t.MyClass.f.__module__ == "test_classes_ext"
    assert t.MyClass().f.__name__ == "f"
    assert t.MyClass().f.__qualname__ == "MyClass.f"
    assert t.MyClass().f.__module__ == "test_classes_ext"
    assert type(t.MyClass.f).__module__ == "nanobind"
    assert type(t.MyClass.f).__name__ == "nb_method"
    assert type(t.MyClass.f).__qualname__ == "nb_method"
    assert t.MyClass.NestedClass.__name__ == "NestedClass"
    assert t.MyClass.NestedClass.__qualname__ == "MyClass.NestedClass"
    assert t.MyClass.NestedClass.__module__ == "test_classes_ext"
    assert t.MyClass.NestedClass.f.__name__ == "f"
    assert t.MyClass.NestedClass.f.__qualname__ == "MyClass.NestedClass.f"
    assert t.MyClass.NestedClass.f.__module__ == "test_classes_ext"


def test18_static_properties():
    assert t.StaticProperties.value == 23
    t.StaticProperties.value += 1
    assert t.StaticProperties.value == 24
    assert t.StaticProperties.get() == 24
    assert t.StaticProperties2.get() == 24
    t.StaticProperties2.value = 50
    assert t.StaticProperties2.get() == 50
    assert t.StaticProperties.get() == 50


@skip_on_pypy
def test19_static_properties_doc():
    import pydoc

    assert "Static property docstring" in pydoc.render_doc(t.StaticProperties2)


def test20_supplement():
    c = t.ClassWithSupplement()
    assert t.check_supplement(c)
    assert not t.check_supplement(t.Struct())


def test21_type_callback():
    o = t.ClassWithLen()
    assert len(o) == 123


def test22_low_level(clean):
    s1, s2, s3, s4 = t.test_lowlevel()
    assert s1.value() == 123 and s2.value() == 0 and s4.value() == 123
    assert s3.s1.value() == 123 and s3.s2.value() == 456
    assert s3.s1 is s4
    del s1
    del s2
    del s3
    collect()
    assert s4.value() == 123
    del s4
    assert_stats(
        value_constructed=3, copy_constructed=1, move_constructed=1, destructed=5
    )


def test23_handle_t(clean):
    assert (
        t.test_handle_t.__doc__
        == "test_handle_t(arg: test_classes_ext.Struct, /) -> object"
    )
    s = t.test_handle_t(t.Struct(5))
    assert s.value() == 5
    del s

    with pytest.raises(TypeError) as excinfo:
        t.test_handle_t("test")
    assert "incompatible function argument" in str(excinfo.value)
    assert_stats(value_constructed=1, destructed=1)


def test24_type_object_t(clean):
    assert (
        t.test_type_object_t.__doc__
        == f"test_type_object_t(arg: {TYPING_TYPE}[test_classes_ext.Struct], /) -> object"
    )

    assert t.test_type_object_t(t.Struct) is t.Struct

    with pytest.raises(TypeError):
        t.test_type_object_t(t.Struct())

    with pytest.raises(TypeError):
        t.test_type_object_t(int)


def test25_none_arg():
    with pytest.raises(TypeError):
        t.none_0(None)
    with pytest.raises(TypeError):
        t.none_1(None)
    with pytest.raises(TypeError):
        t.none_2(arg=None)
    assert t.none_3(None) is True
    assert t.none_4(arg=None) is True
    assert t.none_0.__doc__ == "none_0(arg: test_classes_ext.Struct, /) -> bool"
    assert t.none_1.__doc__ == "none_1(arg: test_classes_ext.Struct) -> bool"
    assert t.none_2.__doc__ == "none_2(arg: test_classes_ext.Struct) -> bool"
    opt_struct = optional("test_classes_ext.Struct")
    assert t.none_3.__doc__ == f"none_3(arg: {opt_struct}) -> bool"
    assert t.none_4.__doc__ == f"none_4(arg: {opt_struct}) -> bool"


def test26_is_final():
    with pytest.raises(TypeError) as excinfo:

        class MyType(t.FinalType):
            pass

    assert "The type 'test_classes_ext.FinalType' prohibits subclassing!" in str(
        excinfo.value
    )


def test27_dynamic_attr(clean):
    l = [None] * 100
    for i in range(100):
        l[i] = t.StructWithAttr(i)

    # Create a big reference cycle..
    for i in range(100):
        l[i].prev = l[i - 1]
        l[i].next = l[i + 1 if i < 99 else 0]
        l[i].t = t.StructWithAttr
        l[i].self = l[i]

    for i in range(100):
        assert l[i].value() == i
        assert l[i].self.value() == i
        assert l[i].prev.value() == (i - 1 if i > 0 else 99)
        assert l[i].next.value() == (i + 1 if i < 99 else 0)

    del l

    assert_stats(value_constructed=100, destructed=100)


def test28_copy_rvp():
    a = t.Struct.create_reference()
    b = t.Struct.create_copy()
    assert a is not b


def test29_pydoc():
    import pydoc

    assert "Some documentation" in pydoc.render_doc(t)


def test30_property_assignment_instance():
    s = t.PairStruct()
    s1 = t.Struct(123)
    s2 = t.Struct(456)
    s.s1 = s1
    s.s2 = s2
    assert s2 is not s.s2 and s1 is not s.s1
    assert s.s1.value() == 123
    assert s.s2.value() == 456
    assert s1.value() == 123
    assert s2.value() == 456


# cpyext reference cycles are not supported, see https://foss.heptapod.net/pypy/pypy/-/issues/3849
@skip_on_pypy()
def test31_cycle():
    a = t.Wrapper()
    a.value = a
    del a
    collect()


def test32_type_checks():
    v1 = 5
    v2 = t.Struct()

    assert t.is_int_1(v1) and not t.is_int_1(v2)
    assert t.is_int_2(v1) and not t.is_int_2(v2)
    assert not t.is_struct(v1) and t.is_struct(v2)


def test33_polymorphic_downcast():
    assert isinstance(t.factory(), t.Base)
    assert isinstance(t.factory_2(), t.Base)
    assert isinstance(t.polymorphic_factory(), t.PolymorphicSubclass)
    assert isinstance(t.polymorphic_factory_2(), t.PolymorphicBase)


def test34_trampoline_optimization():
    class Rufus(t.Dog):
        def __init__(self):
            super().__init__("woof")

        def name(self):
            return "Rufus"

    for i in range(2):
        d1 = t.Dog("woof")
        d2 = Rufus()

        if i == 0:
            assert t.go(d1) == "Dog says woof"
            assert t.go(d2) == "Rufus says woof"

        old = t.Dog.name
        try:
            t.Dog.name = lambda self: "Max"

            assert t.go(d1) == "Dog says woof"
            assert t.go(d2) == "Rufus says woof"
        finally:
            t.Dog.name = old


def test35_method_introspection():
    obj = t.Struct(5)
    m = obj.value
    assert m() == m.__call__() == 5
    assert hash(m) == m.__hash__()
    assert repr(m) == m.__repr__()
    assert "bound_method" in repr(m)
    assert m.__self__ is obj
    assert m.__func__ is t.Struct.value
    # attributes not defined by nb_bound_method are forwarded to nb_method:
    assert m.__name__ == "value"
    assert m.__qualname__ == "Struct.value"
    assert m.__module__ == t.__name__
    assert m.__doc__ == t.Struct.value.__doc__ == "value(self) -> int"


def test38_pickle(clean):
    import pickle

    s = t.Struct(123)
    s2 = pickle.dumps(s, protocol=pickle.HIGHEST_PROTOCOL)
    s3 = pickle.loads(s2)
    assert s.value() == s3.value()
    del s, s3

    assert_stats(value_constructed=1, pickled=1, unpickled=1, destructed=2)


def test39_try_cast(clean):
    s = t.Struct(123)

    assert_stats(value_constructed=1)
    t.reset()

    rv, s2 = t.try_cast_1(s)
    assert rv is True and s2 is not s and s.value() == 123 and s2.value() == 123
    del s2
    assert_stats(
        default_constructed=1, move_constructed=2, copy_assigned=1, destructed=3
    )
    t.reset()

    rv, s2 = t.try_cast_1(None)
    assert rv is False and s2 is not s and s2.value() == 5
    del s2
    assert_stats(
        default_constructed=1, move_constructed=2, copy_assigned=0, destructed=3
    )
    t.reset()

    rv, s2 = t.try_cast_2(s)
    assert rv is True and s2 is not s and s.value() == 123 and s2.value() == 123
    del s2
    assert_stats(
        default_constructed=1, move_constructed=2, copy_assigned=1, destructed=3
    )
    t.reset()

    rv, s2 = t.try_cast_2(None)
    assert rv is False and s2 is not s and s2.value() == 5
    del s2
    assert_stats(
        default_constructed=1, move_constructed=2, copy_assigned=0, destructed=3
    )
    t.reset()

    rv, s2 = t.try_cast_3(s)
    assert rv is True and s2 is s and s.value() == 123
    del s2
    assert_stats()
    t.reset()

    rv, s2 = t.try_cast_3(None)
    assert rv is True and s2 is None
    del s2
    assert_stats(
        default_constructed=0, move_constructed=0, copy_assigned=0, destructed=0
    )
    t.reset()

    rv, s2 = t.try_cast_2(1)
    assert rv is False
    del s2
    assert_stats(default_constructed=1, move_constructed=2, destructed=3)
    t.reset()

    rv, s2 = t.try_cast_3(1)
    assert rv is False and s2 is None
    del s2
    assert_stats()
    t.reset()

    rv, s2 = t.try_cast_4(s)
    assert rv is False and s2 == 0
    rv, s2 = t.try_cast_4(123)
    assert rv is True and s2 == 123
    del s, s2

    assert_stats(destructed=1)


def test40_slots():
    if not hasattr(t, "test_slots"):
        pytest.skip()
    assert t.test_slots() == (True, True, True)


def test41_implicit_conversion_keep_alive():
    # Check that keep_alive references implicitly constructed arguments
    # as opposed to the original function arguments
    collect()
    t.get_destructed()

    a = t.Struct(5)
    b = t.get_incrementing_struct_value(a)
    d1 = t.get_destructed()
    assert b.value() == 106
    del a
    collect()
    d2 = t.get_destructed()
    collect()
    del b
    collect()
    d3 = t.get_destructed()
    assert d1 == []
    assert d2 == [5]
    assert d3 == [106, 6]


def test42_weak_references():
    import weakref
    import gc
    import time

    o = t.StructWithWeakrefs(42)
    w = weakref.ref(o)
    assert w() is o
    del o
    gc.collect()
    gc.collect()
    assert w() is None

    p = t.StructWithWeakrefsAndDynamicAttrs(43)
    p.a_dynamic_attr = 101
    w = weakref.ref(p)
    assert w() is p
    assert w().a_dynamic_attr == 101
    del p
    gc.collect()
    gc.collect()
    assert w() is None


def test43_union():
    u = t.Union()
    u.i = 42
    assert u.i == 42

    u.f = 2.125
    assert u.f == 2.125


def test44_dynamic_attr_has_dict():
    s = t.StructWithAttr(5)
    assert s.__dict__ == {}
    s.a_dynamic_attr = 101
    assert s.__dict__ == {"a_dynamic_attr": 101}


def test45_hidden_base():
    s = t.BoundDerived()
    assert s.value == 10
    s.value = 5
    assert s.prop == 5
    s.prop = 20
    assert s.value == 20
    assert s.get_answer() == 200
    assert s.polymorphic() == 20


def test46_custom_new():
    import gc

    u1 = t.UniqueInt(10)
    assert u1.value() == 10 and u1.lookups() == 1

    u2 = t.UniqueInt(10)
    assert u1 is u2
    assert u1.lookups() == 2

    # test alternate constructor
    assert t.UniqueInt("10") is u1
    assert u1.lookups() == 3

    u3 = t.UniqueInt(20)
    assert u1 is not u3
    assert u3.value() == 20 and u3.lookups() == 1

    del u1
    assert u2.lookups() == 3
    assert u2 is t.UniqueInt(10)
    assert u2.lookups() == 4

    del u2
    gc.collect()
    gc.collect()

    u4 = t.UniqueInt(10)
    assert u4.value() == 10 and u4.lookups() == 1

    # As if unpickling:
    empty = t.UniqueInt.__new__(t.UniqueInt)
    with pytest.warns(RuntimeWarning, match="access an uninitialized instance"):
        with pytest.raises(TypeError):
            empty.value()

    # Make sure pickle support doesn't allow no-args construction by mistake
    with pytest.raises(TypeError):
        t.UniqueInt()

    with pytest.raises(RuntimeError):
        t.UniqueInt.__new__(int)



import test_inter_module_1_ext as t1
import test_inter_module_2_ext as t2
import test_classes_ext as t3
import pytest
from common import xfail_on_pypy_darwin

@xfail_on_pypy_darwin
def test01_inter_module():
    s = t1.create_shared()
    assert t2.check_shared(s)
    with pytest.raises(TypeError) as excinfo:
        assert t3.check_shared(s)
    assert 'incompatible function arguments' in str(excinfo.value)
